from enum import Enum
from functools import partial


class ConstructionInfo:
    class State(Enum):
        DISCARDED = 1
        CONSTRUCTED = 2
        PASSED = 3

    def __init__(self):
        self.n_cluster = None
        self.subject = None
        self.state = None


class ConstLogger:
    def __init__(self, func):
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, learnblock):
        self.log_begin_construction()

        for info, complete_learnblock in self.func(owner, learnblock):
            if info.state == ConstructionInfo.State.PASSED and self.logger:
                self.logger.protocol("State: {:<20}".format("PASSED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            elif info.state == ConstructionInfo.State.DISCARDED and self.logger:
                self.logger.protocol("State: {:<20}".format("DISCARDED"))

            elif info.state == ConstructionInfo.State.CONSTRUCTED and self.logger:
                self.logger.protocol("State: {:<20}".format("CONSTRUCTED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            if complete_learnblock is None: return  # Close Generator

            yield complete_learnblock
            self.log_begin_construction()

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Construction"))

    def log_learnglock_info(self, learnblock, complete_learnblock):
        self.logger.protocol(
            "\t".join(["",
                       "[{:^5}{:^5}]".format(learnblock.n_rows(),
                                             learnblock.n_columns()),
                       str(complete_learnblock.min_timestamp()),
                       str(complete_learnblock.max_timestamp()),
                       "{:<20}".format(str(complete_learnblock.origin)),
                       "{:<20}".format(str(complete_learnblock.relationship)),
                       "{:<20}".format(str(complete_learnblock.subject)),
                       ]))

    def log_passed_info(self):
        pass

    def log_constructed_info(self, logging_info):
        self.logger.protocol("State: {:<20}".format("CONSTRUCTED"))
        self.logger.protocol("{:<20}".format(logging_info.subject))
        self.logger.protocol("{:<20}".format(logging_info.n_cluster))


class ReconstLogger:
    def __init__(self, func):
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, tier, learnblock, which_models=None, meta=None):
        self.log_begin_construction()
        winner = self.func(owner, tier, learnblock, which_models, meta)
        self.log_winner(winner)
        return winner

    def log_winner(self, winner):
        if self.logger:
            self.logger.protocol("Subject: {:<20}".format(str(winner.subject)))
            self.logger.protocol("Accuracy: {:<20}".format(winner.accuracy))
            self.logger.protocol("Reliability: {:<20}".format(winner.reliability))

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Reconstruction"))


class DeconstInfo:
    def __init__(self):
        self.relationship = None
        self.relative = None
        self.pragmatic = None


class DeconstLoggger:
    def __init__(self, func):
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, pragmatic, relative, block):
        logging_info, success = self.func(owner, pragmatic, relative, block)
        self.log(logging_info)
        return success

    def log(self, info):
        self.logger.protocol(
            "{:*^100}".format("Deconstruction {}".format(info.relationship)))
        self.logger.protocol("Pragmatic: {:<20}".format(str(info.pragmatic)))
        self.logger.protocol("Relative: {:<20}".format(str(info.relative)))


